package nq.utils.task {

    import flash.events.Event;
    import flash.events.EventDispatcher;
    import nq.core.IDestroyable;

    [Event(name = "queueComplete", type = "nq.utils.task.TaskQueueEvent")]
	[Event(name = "beforeTaskStart", type = "nq.utils.task.TaskQueueEvent")]
	[Event(name = "taskComplete", type = "nq.utils.task.TaskQueueEvent")]
	[Event(name = "taskFail", type = "nq.utils.task.TaskQueueEvent")]

	/**
	 * Многопоточный испольнитель заданий
	 *
	 * @author gloomyBrain
	 */
	public final class TaskQueue extends EventDispatcher implements IDestroyable {

		private var _queuedTasks:Vector.<ITask>;
		private var _runningTasks:Vector.<IAsyncTask>;

		private var _running:Boolean;

		private var _maxThreads:uint;
		private var _usePriority:Boolean;

		/**
		 * @param	maxThreads	Максимальное число одновременно
		 * выполняющихся асинхронных заданий
		 * @param	usePriority	Использовать ли приоритеты
		 */
		public function TaskQueue(maxThreads:uint = 1, usePriority:Boolean = false) {

			super();

			_maxThreads = (maxThreads > 0) ? maxThreads : 1;
			_usePriority = usePriority;

			_running = false;

			_queuedTasks = new Vector.<ITask>();
			_runningTasks = new Vector.<IAsyncTask>();
		}

		/**
		 * Запустить выполнение очереди заданий
		 */
		public function start():void {

			if (_running) return;

			_running = true;
			this.checkQueue();
		}

		/**
		 * Остановить выполнение очереди заданий
		 */
		public function stop():void {

			if (!_running) return;
			_running = false;
		}

		/**
		 * Немедленно остановить очередь и все выполняемые задания
		 */
		public function stopImmediately():void {

			var len:int = _runningTasks.length;
			while (len--)
			{
				this.deactivateAsyncTask(_runningTasks[len]);
				_runningTasks[len].stop();
			}

			_runningTasks.length = 0;

			this.stop();

		}

		/**
		 * Удалить все отложенные задания
		 */
		public function clear():void {
			_queuedTasks.length = 0;
		}

		/**
		 * Добавить задание в очередь
		 *
		 * @param	task	Добавляемое задание
		 */
		public function addTask(task:ITask):void {
			if (!task) throw new ArgumentError("Параметр не должен быть нулевым!");

			if (
				!_usePriority ||
				_queuedTasks.length == 0 ||
				_queuedTasks[_queuedTasks.length - 1].priority > task.priority
			) {

				_queuedTasks.push(task);

			}else {

				var index:int = _queuedTasks.length;

				while (index--) {

					if (_queuedTasks[index].priority < task.priority) continue;

					_queuedTasks.splice(index + 1, 0, task);
					return;

				}

				_queuedTasks.unshift(task);

			}
		}

		/**
		 * Убрать задание из очереди (если задание выполняется, оно
		 * будет остановлено принудительно)
		 *
		 * @param	task	Удаляемое задание
		 */
		public function removeTask(task:ITask):void {
			if (!task) throw new ArgumentError("Параметр не должен быть нулевым!");

			// Ищем задание в списке ожидающих
			var index:int = _queuedTasks.indexOf(task);
			if (index != -1) {

				_queuedTasks.splice(index, 1);
				return;
			}

			// Если задание не асинхронное, то оно не может
			// попасть в очередь выполняемых
			var asyncTask:IAsyncTask = task as IAsyncTask;
			if (!asyncTask) return;

			index = _runningTasks.indexOf(asyncTask);
			if (index == -1) return;

			this.deactivateAsyncTask(asyncTask);
			_runningTasks.splice(index, 1);
			asyncTask.stop();

			this.checkQueue();
		}

        /**
         * Возвращает true если таска запущена
         * @param task задание
         * @return true или false
         */
        public function isTaskRunning(task:IAsyncTask):Boolean
        {
            return _runningTasks.indexOf(task) != -1;
        }

		private function checkQueue():void {

			// пока есть ожидающие задания и выполняемых заданий меньше
			// чем число потоков - набиваем буфер и стартуем задания
			while (_running && _queuedTasks.length && (_runningTasks.length < _maxThreads)) {
				this.startNextTask();
			}

			// Если нет выполняемых заданий, значит очередь завершена
			if (_running && _runningTasks.length == 0) {

				_running = false;
				super.dispatchEvent(new TaskQueueEvent(TaskQueueEvent.QUEUE_COMPLETE));
			}

		}

		private function startNextTask():void {

			var task:ITask = _queuedTasks[0];

			super.dispatchEvent(new TaskQueueEvent(TaskQueueEvent.BEFORE_TASK_START, task));
			if (!(_running && _queuedTasks.length && (task == _queuedTasks[0]))) return;

			_queuedTasks.shift();

			var asyncTask:IAsyncTask = task as IAsyncTask;
			if (!asyncTask) {

				task.start();

				super.dispatchEvent(new TaskQueueEvent(TaskQueueEvent.TASK_COMPLETE, task));
				return;
			}

			asyncTask.addEventListener(Event.COMPLETE, onAsyncTaskComplete);
			asyncTask.addEventListener(Event.CANCEL, onAsyncTaskCancel);

			_runningTasks.push(asyncTask);

			asyncTask.start();
		}

		private function removeAsyncTask(asyncTask:IAsyncTask):void {

			this.deactivateAsyncTask(asyncTask);

			var index:int = _runningTasks.indexOf(asyncTask);
			_runningTasks.splice(index, 1);
		}

		private function deactivateAsyncTask(asyncTask:IAsyncTask):void {

			asyncTask.removeEventListener(Event.COMPLETE, onAsyncTaskComplete);
			asyncTask.removeEventListener(Event.CANCEL, onAsyncTaskCancel);
		}

		private function onAsyncTaskComplete(event:Event):void {

			var asyncTask:IAsyncTask = event.currentTarget as IAsyncTask;
			this.removeAsyncTask(asyncTask);

			super.dispatchEvent(new TaskQueueEvent(TaskQueueEvent.TASK_COMPLETE, asyncTask));

			this.checkQueue();
		}

		private function onAsyncTaskCancel(event:Event):void {

			var asyncTask:IAsyncTask = event.currentTarget as IAsyncTask;
			this.removeAsyncTask(asyncTask);

			super.dispatchEvent(new TaskQueueEvent(TaskQueueEvent.TASK_FAIL, asyncTask));

			this.checkQueue();
		}

		/**
		 * Число заданий, выполняемых в данный момент
		 */
		public function get runningTasks():uint {
			return _runningTasks.length;
		}

		/**
		 * Число ожидающих заданий в очереди
		 */
		public function get queuedTasks():uint {
			return _queuedTasks.length;
		}

		/**
		 * Исполняется ли очередь в данный момент
		 */
		public function get isRunning():Boolean {
			return _running;
		}


		/* INTERFACE nq.core.IDestroyable */

		/**
		 * @inheritDoc
		 */
		public function destroy():void {

			this.stopImmediately();
			this.clear();

			_queuedTasks = null;
			_runningTasks = null;

		}

        /**
         * Все таски этой очереди
         */
        public function get tasks():Vector.<ITask>
        {
            return _queuedTasks.concat(_runningTasks);
        }
	}
}